"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const Credentials_1 = __importDefault(require("../credentials/Credentials"));
const Entry_1 = __importDefault(require("../core/Entry"));
const Vault_1 = __importDefault(require("../core/Vault"));
const appEnv_1 = require("../env/appEnv");
const uuid_1 = require("../tools/uuid");
const attachments_1 = require("../tools/attachments");
const ATTACHMENTS_KEY_LENGTH = 48;
/**
 * Attachment manager
 * @memberof module:Buttercup
 */
class AttachmentManager {
    /**
     * Get a new attachment ID
     * @static
     * @memberof AttachmentManager
     */
    static newAttachmentID() {
        return (0, uuid_1.generateUUID)();
    }
    /**
     * Constructor for new attachment managers
     * @param vaultSource The vault source to attach to. This is
     *  normally set by the VaultSource automatically when unlocking a source.
     */
    constructor(vaultSource) {
        this._source = vaultSource;
    }
    /**
     * Get an attachment's data
     * @param entry Entry instance the attachment can be found on
     * @param attachmentID The attachment ID
     * @throws {Error} Throws if the attachment isn't found
     * @memberof AttachmentManager
     */
    getAttachment(entry, attachmentID) {
        return __awaiter(this, void 0, void 0, function* () {
            this._checkAttachmentSupport();
            const details = yield this.getAttachmentDetails(entry, attachmentID);
            if (!details) {
                throw new Error(`Attachment not available: ${attachmentID}`);
            }
            const credentials = yield this._getAttachmentsCredentials();
            const data = yield this._source._datasource.getAttachment(this._source.vault.id, attachmentID);
            return (0, attachments_1.decryptAttachment)(data, credentials);
        });
    }
    /**
     * Get an attachment's details
     * @param entry Entry instance the attachment can be found on
     * @param attachmentID The attachment ID
     * @returns The details or null if the attachment isn't found
     * @memberof AttachmentManager
     */
    getAttachmentDetails(entry, attachmentID) {
        return __awaiter(this, void 0, void 0, function* () {
            this._checkAttachmentSupport();
            const attributeKey = `${Entry_1.default.Attributes.AttachmentPrefix}${attachmentID}`;
            const payloadRaw = entry.getAttribute(attributeKey);
            return payloadRaw ? JSON.parse(payloadRaw) : null;
        });
    }
    /**
     * List all attachments
     * @param entry Entry instance the attachment can be found on
     * @memberof AttachmentManager
     */
    listAttachments(entry) {
        return __awaiter(this, void 0, void 0, function* () {
            this._checkAttachmentSupport();
            const attributes = entry.getAttribute();
            const attachmentKeys = Object.keys(attributes).filter(key => key.indexOf(Entry_1.default.Attributes.AttachmentPrefix) === 0);
            return attachmentKeys.map(key => JSON.parse(attributes[key]));
        });
    }
    /**
     * Remove an attachment, deleting the file and removing it from
     *  the entry
     * @param entry Entry instance the attachment can be found on
     * @param attachmentID The attachment ID
     * @memberof AttachmentManager
     */
    removeAttachment(entry, attachmentID) {
        return __awaiter(this, void 0, void 0, function* () {
            this._checkAttachmentSupport();
            const attributeKey = `${Entry_1.default.Attributes.AttachmentPrefix}${attachmentID}`;
            // Remove data
            yield this._source._datasource.removeAttachment(this._source.vault.id, attachmentID);
            // Remove in entry
            entry.deleteAttribute(attributeKey);
        });
    }
    /**
     * Write an attachment to an entry.
     * Creates a new attachment or updates an existing one, by attachment ID
     * @param entry Entry instance the attachment can be found on
     * @param attachmentID The attachment ID
     * @param attachmentData The attachment's data
     * @param name The file name
     * @param type The MIME type
     * @param timestamp Optional timestamp override for the creation of the
     *  attachment
     * @memberof AttachmentManager
     */
    setAttachment(entry, attachmentID, attachmentData, name, type, timestamp = new Date()) {
        return __awaiter(this, void 0, void 0, function* () {
            this._checkAttachmentSupport();
            if (!name || !type) {
                throw new Error(`Attachment properties required: name/type => ${name}/${type}`);
            }
            const attributeKey = `${Entry_1.default.Attributes.AttachmentPrefix}${attachmentID}`;
            // Check if it already exists
            const existingDetails = yield this.getAttachmentDetails(entry, attachmentID);
            // Get credentials
            const credentials = yield this._getAttachmentsCredentials();
            // Encrypt it
            const encryptedAttachmentData = yield (0, attachments_1.encryptAttachment)(attachmentData, credentials);
            const encryptedSize = (0, attachments_1.getBufferSize)(encryptedAttachmentData);
            const originalSize = (0, attachments_1.getBufferSize)(attachmentData);
            // Calculate if it can fit in storage
            const sizeIncrease = existingDetails
                ? Math.max(0, encryptedSize - existingDetails.sizeEncrypted)
                : encryptedSize;
            const spaceAvailable = yield this._source._datasource.getAvailableStorage();
            if (spaceAvailable !== null && sizeIncrease > spaceAvailable) {
                throw new Error(`Not enough space to update attachment: needed = ${sizeIncrease} B, available = ${spaceAvailable} B`);
            }
            // Create payload
            const now = timestamp.toUTCString();
            const payload = {
                id: attachmentID,
                name,
                type,
                sizeOriginal: originalSize,
                sizeEncrypted: encryptedSize,
                created: now,
                updated: now
            };
            // Write attachment
            yield this._source._datasource.putAttachment(this._source.vault.id, attachmentID, encryptedAttachmentData, payload);
            // Set in entry
            entry.setAttribute(attributeKey, JSON.stringify(payload));
        });
    }
    _checkAttachmentSupport() {
        if (!this._source.supportsAttachments()) {
            throw new Error(`Attachments not supported on source: ${this._source.id}`);
        }
    }
    _getAttachmentsCredentials() {
        return __awaiter(this, void 0, void 0, function* () {
            const { vault } = this._source;
            let key = vault.getAttribute(Vault_1.default.Attribute.AttachmentsKey);
            if (!key) {
                // Create key for first-time use
                const generateRandomString = (0, appEnv_1.getSharedAppEnv)().getProperty("crypto/v1/randomString");
                key = yield generateRandomString(ATTACHMENTS_KEY_LENGTH);
                vault.setAttribute(Vault_1.default.Attribute.AttachmentsKey, key);
                // Save the key
                yield this._source.save();
            }
            return Credentials_1.default.fromPassword(key);
        });
    }
}
exports.default = AttachmentManager;
