"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const eventemitter3_1 = __importDefault(require("eventemitter3"));
const layerr_1 = require("layerr");
const is_promise_1 = __importDefault(require("is-promise"));
const channel_queue_1 = require("@buttercup/channel-queue");
const MemoryStorageInterface_1 = __importDefault(require("../storage/MemoryStorageInterface"));
const VaultSource_1 = __importDefault(require("./VaultSource"));
const DEFAULT_AUTO_UPDATE_DELAY = 1000 * 60 * 2.5; // 2.5 mins
const NOOP = () => { };
const STORAGE_KEY_PREFIX = "bcup_vaultmgr_";
const STORAGE_KEY_PREFIX_TEST = /^bcup_vaultmgr_[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}$/;
/**
 * Vault manager, to manage vault sources and their vaults
 * @augments EventEmitter
 * @memberof module:Buttercup
 */
class VaultManager extends eventemitter3_1.default {
    /**
     * Construct a new VaultManager
     * @param opts Configuration options
     */
    constructor(opts = {}) {
        super();
        this._autoUpdateTimer = null;
        this._initialised = false;
        this._queue = new channel_queue_1.ChannelQueue();
        this._sources = [];
        const { autoUpdate = true, autoUpdateDelay = DEFAULT_AUTO_UPDATE_DELAY, cacheStorage = new MemoryStorageInterface_1.default(), sourceStorage = new MemoryStorageInterface_1.default() } = opts;
        this._cacheStorage = cacheStorage;
        this._sourceStorage = sourceStorage;
        this._autoUpdateEnabled = autoUpdate;
        this._autoUpdateDelay = autoUpdateDelay;
    }
    /**
     * Array of vault sources
     * @readonly
     * @memberof VaultManager
     */
    get sources() {
        return [...this._sources];
    }
    /**
     * Array of unlocked vault sources
     * @readonly
     * @memberof VaultManager
     */
    get unlockedSources() {
        return this._sources.filter(source => source.status === VaultSource_1.default.STATUS_UNLOCKED);
    }
    /**
     * @typedef {Object} VaultManagerAddSourceOptions
     * @property {Number=} order Optional order override
     */
    /**
     * Add a VaultSource to the VaultManager
     * - The vault manager will then provide a management
     * platform for the source, including storage access and
     * event aggregation.
     * @param source The source to add
     * @param opts Options for adding the
     *  source
     * @memberof VaultManager
     */
    addSource(source, opts = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            const { order: orderOverride } = opts;
            const existing = this._sources.find(src => src.id === source.id);
            if (existing)
                return;
            yield this.enqueueStateChange(() => __awaiter(this, void 0, void 0, function* () {
                source._vaultManager = this;
                const newOrder = typeof orderOverride === "number" ? orderOverride : this.getNextOrder();
                this._sources.push(source);
                // Configure the order
                source.order = newOrder;
                // Attach event listeners
                source.on("updated", () => this.dehydrateSource(source));
                source.on("updated", () => this.emit("sourcesUpdated"));
                source.on("locked", () => this.emit("sourcesUpdated"));
                source.on("unlocked", () => this.emit("sourcesUpdated"));
                const dehydratedString = yield source.dehydrate();
                yield this._storeDehydratedSource(source.id, dehydratedString);
                this.emit("sourcesUpdated");
            }));
        });
    }
    /**
     * Dehydrate all sources and write them to storage
     * @returns A promise that resolves once all sources have been dehydrated
     * @memberof VaultManager
     */
    dehydrate() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.enqueueStateChange(() => Promise.all(this._sources.map(source => this.dehydrateSource(source))));
        });
    }
    /**
     * Dehydrate a single archive source
     * @param sourceID The ID of the source
     * @returns A promise that resolves once the source has been dehydrated
     * @memberof VaultManager
     */
    dehydrateSource(sourceOrSourceID) {
        return __awaiter(this, void 0, void 0, function* () {
            const source = typeof sourceOrSourceID === "string" ? this.getSourceForID(sourceOrSourceID) : sourceOrSourceID;
            const dehydratedString = yield source.dehydrate();
            yield this._storeDehydratedSource(source.id, dehydratedString);
        });
    }
    /**
     * Enqueue an asychronous change of state
     * @param cb Callback to enqueue
     * @memberof VaultManager
     */
    enqueueStateChange(cb) {
        return this._queue.channel("state").enqueue(cb);
    }
    /**
     * Get the next viable order number for a new source
     * @returns The new order
     * @memberof VaultManager
     */
    getNextOrder() {
        return Math.max(...this._sources.map(source => source._order), -1) + 1;
    }
    /**
     * Get a source for an ID
     * @param sourceID The source ID
     * @returns The source with the matching ID or null if not found
     * @memberof VaultManager
     */
    getSourceForID(sourceID) {
        const source = this._sources.find(target => target.id && target.id === sourceID);
        return source || null;
    }
    /**
     * Get an array of sources that can be updated
     * @memberof VaultManager
     */
    getUpdateableSources() {
        return this._sources.filter(source => source.canBeUpdated());
    }
    /**
     * Initialise the vault manager
     * @memberof VaultManager
     */
    initialise() {
        if (this._initialised)
            return;
        this._startAutoUpdateTimer();
        this._initialised = true;
    }
    /**
     * Wait for and interrupt state changes when auto-update is running
     * @param cb The callback to execute during the auto-update interruption
     * @returns A promise that resolves when ready
     * @memberof VaultManager
     * @example
     *  vaultManager.interruptAutoUpdate(() => {
     *      // Do something with auto-updating paused
     *  });
     */
    interruptAutoUpdate(cb) {
        return this.enqueueStateChange(NOOP).then(() => this._queue.channel("autoUpdateInterrupt").enqueue(() => {
            const enabled = this._autoUpdateEnabled;
            const delay = this._autoUpdateDelay;
            this.toggleAutoUpdating(false);
            const restoreAutoUpdating = () => {
                this.toggleAutoUpdating(enabled, delay);
            };
            let retVal;
            try {
                retVal = cb();
            }
            catch (err) {
                // Callback died, so restore and throw
                restoreAutoUpdating();
                throw err;
            }
            if (!(0, is_promise_1.default)(retVal)) {
                // No promise, so restore immediately:
                restoreAutoUpdating();
                return retVal;
            }
            return retVal
                .then(res => {
                // Wait until after the promise has completed
                // or failed:
                restoreAutoUpdating();
                return res;
            })
                .catch(err => {
                restoreAutoUpdating();
                throw err;
            });
        }));
    }
    /**
     * Rehydrate sources from storage
     * @returns A promise that resolves once rehydration has completed
     * @memberof VaultManager
     * @throws {Layerr} Rejects if rehydrating from storage fails
     */
    rehydrate() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this._migrateLegacyVaults();
            const storageKeys = yield this._sourceStorage.getAllKeys();
            yield Promise.all(storageKeys
                .filter(key => STORAGE_KEY_PREFIX_TEST.test(key))
                .map((key) => __awaiter(this, void 0, void 0, function* () {
                const dehydratedSource = yield this._sourceStorage.getValue(key);
                const source = VaultSource_1.default.rehydrate(dehydratedSource);
                yield this.addSource(source, {
                    order: source.order
                });
            })));
            this.reorderSources();
            this.emit("sourcesUpdated");
        });
    }
    /**
     * Remove a source from the storage
     * @param sourceID The ID of the source to remove
     * @returns A promise that resolves once the source has been removed
     * @memberof VaultManager
     */
    removeSource(sourceID) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.enqueueStateChange(() => __awaiter(this, void 0, void 0, function* () {
                const sourceIndex = this._sources.findIndex(source => source.id === sourceID);
                if (sourceIndex === -1) {
                    throw new layerr_1.Layerr(`Failed removing source: No source found for ID: ${sourceID}`);
                }
                const source = this.sources[sourceIndex];
                source.removeAllListeners();
                this._sources.splice(sourceIndex, 1);
                this.emit("sourcesUpdated");
                yield this._sourceStorage.removeKey(`${STORAGE_KEY_PREFIX}${sourceID}`);
            }));
        });
    }
    /**
     * Reorder a source
     * @param sourceID The ID of the source to reorder
     * @param position The 0-based position to move the source to
     * @memberof VaultManager
     * @throws {Layerr} Throws if no source is found
     */
    reorderSource(sourceID, position) {
        return __awaiter(this, void 0, void 0, function* () {
            const source = this.getSourceForID(sourceID);
            if (!source) {
                throw new layerr_1.Layerr(`Failed reordering source: No source found for ID: ${sourceID}`);
            }
            if (position === source._order) {
                return;
            }
            const originalOrder = source._order;
            source._order = position;
            const movingUp = position < originalOrder;
            this.sources.forEach(otherSource => {
                if (otherSource.id !== sourceID) {
                    if (movingUp && otherSource._order >= position) {
                        otherSource._order += 1;
                    }
                    else if (!movingUp && otherSource._order <= position) {
                        otherSource._order -= 1;
                    }
                }
            });
            this.reorderSources();
            // Sync orders to storage
            yield this.dehydrateSource(source);
        });
    }
    /**
     * Reorder all sources
     * @memberof VaultManager
     */
    reorderSources() {
        this._sources.sort((sourceA, sourceB) => {
            if (sourceA._order > sourceB._order) {
                return 1;
            }
            else if (sourceB._order > sourceA._order) {
                return -1;
            }
            return 0;
        });
        this._sources.forEach((source, index) => {
            source._order = index;
        });
        this.emit("sourcesUpdated");
    }
    /**
     * Toggle auto updating of sources
     * @param enable Enable or disable auto updating. Leave empty
     *  to invert the setting
     * @param delay Milliseconds between updates
     * @memberof VaultManager
     */
    toggleAutoUpdating(enable = !this._autoUpdateEnabled, delay = this._autoUpdateDelay) {
        if (enable) {
            this._autoUpdateDelay = delay;
            this._startAutoUpdateTimer();
        }
        else {
            this._stopAutoUpdateTimer();
        }
    }
    _autoUpdateSources() {
        this.emit("autoUpdateStart");
        return this.enqueueStateChange(() => __awaiter(this, void 0, void 0, function* () {
            const updateableSources = this.getUpdateableSources();
            if (updateableSources.length <= 0) {
                return;
            }
            yield Promise.all(updateableSources.map(source => source.update().catch(err => {
                // we don't return auto-update errors, but emit them
                console.error(`Failed auto-updating source: ${source.id}`, err);
                this.emit("autoUpdateFailed", { source, error: err });
            })));
            this.emit("autoUpdateStop");
        }));
    }
    _migrateLegacyVaults() {
        return __awaiter(this, void 0, void 0, function* () {
            const legacyPrefix = "bcup_archivemgr_";
            const storageKeys = yield this._sourceStorage.getAllKeys();
            yield Promise.all(storageKeys
                .filter(key => key && key.indexOf(legacyPrefix) === 0)
                .map((key) => __awaiter(this, void 0, void 0, function* () {
                const value = yield this._sourceStorage.getValue(key);
                if (!value)
                    return;
                const newKey = key.replace(legacyPrefix, STORAGE_KEY_PREFIX);
                yield this._sourceStorage.setValue(newKey, value);
                yield this._sourceStorage.removeKey(key);
            })));
        });
    }
    _startAutoUpdateTimer() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.enqueueStateChange(() => {
                clearTimeout(this._autoUpdateTimer);
                this._autoUpdateTimer = setTimeout(() => {
                    this._autoUpdateSources().then(() => {
                        if (this._autoUpdateEnabled) {
                            this._startAutoUpdateTimer();
                        }
                    });
                }, this._autoUpdateDelay);
            });
        });
    }
    _stopAutoUpdateTimer() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.enqueueStateChange(() => {
                clearTimeout(this._autoUpdateTimer);
                this._autoUpdateTimer = null;
            });
        });
    }
    _storeDehydratedSource(id, dehydratedSource) {
        return this._sourceStorage.setValue(`${STORAGE_KEY_PREFIX}${id}`, dehydratedSource);
    }
}
exports.default = VaultManager;
/**
 * Key prefix for stored vaults
 * @static
 * @memberof VaultManager
 */
VaultManager.STORAGE_KEY_PREFIX = STORAGE_KEY_PREFIX;
