"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const eventemitter3_1 = __importDefault(require("eventemitter3"));
const channel_queue_1 = require("@buttercup/channel-queue");
const layerr_1 = require("layerr");
const Vault_1 = __importDefault(require("./Vault"));
const Credentials_1 = __importDefault(require("../credentials/Credentials"));
const channel_1 = require("../credentials/channel");
const encoding_1 = require("../tools/encoding");
const vaultManagement_1 = require("../tools/vaultManagement");
const register_1 = require("../datasources/register");
const vault_1 = require("../insight/vault");
const AttachmentManager_1 = __importDefault(require("../attachments/AttachmentManager"));
const types_1 = require("../types");
const conversion_1 = require("../io/formatB/conversion");
const index_common_1 = require("../index.common");
const COLOUR_TEST = /^#([a-f0-9]{3}|[a-f0-9]{6})$/i;
const DEFAULT_COLOUR = "#000000";
const DEFAULT_ORDER = 1000;
function processDehydratedCredentials(credentialsString, masterPassword) {
    if (/^v1\n/.test(credentialsString)) {
        const [, sourceCredStr] = credentialsString.split("\n");
        return Credentials_1.default.fromSecureString(sourceCredStr, masterPassword);
    }
    return Credentials_1.default.fromSecureString(credentialsString, masterPassword);
}
/**
 * Vault source class for managing a single vault
 * within a vault manager
 * @augments EventEmitter
 * @memberof module:Buttercup
 */
class VaultSource extends eventemitter3_1.default {
    /**
     * Rehydrate the vault source from a dehydrated state
     * @param dehydratedString The dehydrated form of the vault source
     * @returns A rehydrated instance
     * @memberof VaultSource
     * @static
     */
    static rehydrate(dehydratedString) {
        const target = JSON.parse(dehydratedString);
        let credentials = target.credentials;
        if (target.v !== 2) {
            const { sourceCredentials, archiveCredentials } = target;
            if (!sourceCredentials || !archiveCredentials) {
                throw new Error("Invalid legacy vault state: missing credentials");
            }
            credentials = `v1\n${sourceCredentials}\n${archiveCredentials}`;
        }
        const { id, name, type, colour = DEFAULT_COLOUR, order = DEFAULT_ORDER, meta = {} } = target;
        return new VaultSource(name, type, credentials, {
            id,
            colour,
            order,
            meta
        });
    }
    constructor(name, type, credentialsString, config = {}) {
        super();
        this._attachmentManager = null;
        this._datasource = null;
        this._shares = [];
        this._vault = null;
        this._vaultManager = null;
        const { colour = DEFAULT_COLOUR, id = (0, encoding_1.getUniqueID)(), order = DEFAULT_ORDER, meta = {} } = config;
        // Queue for managing state transitions
        this._queue = new channel_queue_1.ChannelQueue();
        // Credentials state and status go hand-in-hand:
        //  - Locked = credentials string
        //  - Unlocked = credentials instance
        this._credentials = credentialsString;
        this._status = VaultSource.STATUS_LOCKED;
        // Set other configuration items to properties
        this._id = id;
        this._name = name;
        this._type = type;
        this._colour = colour;
        this._order = order;
        this._meta = meta;
    }
    /**
     * The attachment manager
     * @memberof VaultSource
     * @readonly
     */
    get attachmentManager() {
        return this._attachmentManager;
    }
    /**
     * Source colour
     * @memberof VaultSource
     */
    get colour() {
        return this._colour;
    }
    /**
     * Source ID
     * @memberof VaultSource
     * @readonly
     */
    get id() {
        return this._id;
    }
    /**
     * Meta data
     * @memberof VaultSource
     * @readonly
     */
    get meta() {
        return Object.assign({}, this._meta);
    }
    /**
     * Source name
     * @memberof VaultSource
     * @readonly
     */
    get name() {
        return this._name;
    }
    /**
     * The vault order on a vault management instance
     * @memberof VaultSource
     * @readonly
     */
    get order() {
        return this._order;
    }
    /**
     * Source status
     * @memberof VaultSource
     * @readonly
     */
    get status() {
        return this._status;
    }
    /**
     * The datasource type
     * @memberof VaultSource
     * @readonly
     */
    get type() {
        return this._type;
    }
    /**
     * Vault reference
     * @memberof VaultSource
     * @readonly
     */
    get vault() {
        return this._vault;
    }
    set colour(newColour) {
        if (COLOUR_TEST.test(newColour) !== true) {
            throw new layerr_1.Layerr(`Failed setting colour: Invalid format (expected hex): ${newColour}`);
        }
        this._colour = newColour;
        this.emit("updated");
    }
    set order(newOrder) {
        if (isNaN(newOrder) || typeof newOrder !== "number" || newOrder < 0) {
            throw new layerr_1.Layerr(`Failed setting order: Order must be greater than or equal to 0: ${newOrder}`);
        }
        this._order = newOrder;
        this.emit("updated");
    }
    /**
     * Change the master vault password
     * @param oldPassword The original/current password
     * @param newPassword The new password to change to
     * @param meta Optional metadata
     * @memberof VaultSource
     */
    changeMasterPassword(oldPassword, newPassword, meta = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            if (oldPassword === newPassword) {
                throw new Error("New password cannot be the same as the previous one");
            }
            else if (!newPassword) {
                throw new Error("New password must be specified");
            }
            const datasourceSupportsChange = this._datasource.supportsPasswordChange();
            const newMasterCreds = new Credentials_1.default(meta, newPassword);
            let wasLocked = false;
            if (this.status !== VaultSource.STATUS_UNLOCKED) {
                wasLocked = true;
                // Locked, so unlock
                yield this.unlock(Credentials_1.default.fromPassword(oldPassword));
            }
            else {
                // Unlocked, so check password..
                const credentials = (0, channel_1.getCredentials)(this._credentials.id);
                if (credentials.masterPassword !== oldPassword) {
                    throw new Error("Old password does not match current unlocked instance value");
                }
                // ..and then update
                yield this.update();
            }
            // Check datasource is ready
            if (datasourceSupportsChange) {
                const isReady = yield this._datasource.changePassword((0, register_1.prepareDatasourceCredentials)(newMasterCreds, this._datasource.type), 
                /* preflight: */ true);
                if (!isReady) {
                    throw new Error("Datasource not capable of changing password at this time");
                }
            }
            // Clear offline cache
            yield (0, vaultManagement_1.storeSourceOfflineCopy)(this._vaultManager._cacheStorage, this.id, null);
            // Change password
            const newCredentials = Credentials_1.default.fromCredentials(this._credentials, oldPassword);
            const newCreds = (0, channel_1.getCredentials)(newCredentials.id);
            newCreds.masterPassword = newPassword;
            yield this._updateVaultCredentials(newCredentials);
            // Re-lock if it was locked earlier
            if (wasLocked) {
                yield this.lock();
            }
            // Change remote if supported
            if (datasourceSupportsChange) {
                yield this._datasource.changePassword((0, register_1.prepareDatasourceCredentials)(newMasterCreds, this._datasource.type), 
                /* preflight: */ false);
            }
            this.emit("passwordChanged");
            this.emit("updated");
        });
    }
    /**
     * Check if the vault source can be updated
     * @memberof VaultSource
     */
    canBeUpdated() {
        return this.status === VaultSource.STATUS_UNLOCKED && this._vault.format.dirty === false;
    }
    /**
     * Check if the source has an offline copy
     * @returns A promise which resolves with whether an offline
     *  copy is available or not
     * @memberof VaultSource
     */
    checkOfflineCopy() {
        return (0, vaultManagement_1.sourceHasOfflineCopy)(this._vaultManager._cacheStorage, this.id);
    }
    /**
     * Convert vault to a new format
     * @param targetFormat The target format to convert to
     * @memberof VaultSource
     */
    convert(targetFormat) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.status !== VaultSource.STATUS_UNLOCKED) {
                throw new layerr_1.Layerr(`Failed converting source: Source not unlocked (${this.status}): ${this.id}`);
            }
            if (this.vault.format.getFormat().getFormatID() !== types_1.VaultFormatID.A) {
                throw new layerr_1.Layerr(`Failed converting source: Source not in expected format: ${this.id}`);
            }
            if (targetFormat !== types_1.VaultFormatID.B) {
                throw new layerr_1.Layerr(`Failed converting source: Target format not valid: ${this.id}`);
            }
            const formatBSource = (0, conversion_1.convertFormatAVault)(this.vault.format.source);
            const format = new index_common_1.VaultFormatB(formatBSource);
            this._vault = new Vault_1.default(format);
        });
    }
    /**
     * Dehydrate the source to a JSON string, ready for storage
     * @memberof VaultSource
     */
    dehydrate() {
        return this._enqueueStateChange(() => __awaiter(this, void 0, void 0, function* () {
            const payload = {
                v: 2,
                id: this.id,
                name: this.name,
                type: this.type,
                status: VaultSource.STATUS_LOCKED,
                colour: this.colour,
                order: this.order,
                meta: this.meta,
                credentials: null
            };
            if (this.status === VaultSource.STATUS_PENDING) {
                throw new layerr_1.Layerr(`Failed dehydrating source: Source in pending state: ${this.id}`);
            }
            else if (this.status === VaultSource.STATUS_LOCKED) {
                payload.credentials = this._credentials;
            }
            else {
                payload.credentials = yield this._credentials.toSecureString();
            }
            return JSON.stringify(payload);
        }));
    }
    /**
     * Get offline content, if it exists
     * @returns A promise a resolves with the content, or null
     *  if it doesn't exist
     * @memberof VaultSource
     */
    getOfflineContent() {
        return __awaiter(this, void 0, void 0, function* () {
            const hasContent = yield this.checkOfflineCopy();
            if (hasContent) {
                return (0, vaultManagement_1.getSourceOfflineArchive)(this._vaultManager._cacheStorage, this.id);
            }
            return null;
        });
    }
    /**
     * Detect whether the local archives (in memory) differ from their remote copies
     * Fetches the remote copies from their datasources and detects differences between
     * them and their local counterparts. Does not change/update the local items.
     * @returns A promise that resolves with a boolean - true if
     *      there are differences, false if there is not
     * @memberof VaultSource
     */
    localDiffersFromRemote() {
        if (this.status !== VaultSource.STATUS_UNLOCKED) {
            return Promise.reject(new layerr_1.Layerr(`Failed diffing source: Source not unlocked (${this.status}): ${this.id}`));
        }
        if (typeof this._datasource.localDiffersFromRemote === "function") {
            return this._datasource.localDiffersFromRemote((0, register_1.prepareDatasourceCredentials)(this._credentials, this._datasource.type), this.vault.format.history);
        }
        if (this._datasource.type !== "text") {
            // Only clear if not a TextDatasource
            this._datasource.setContent("");
        }
        return this._datasource
            .load((0, register_1.prepareDatasourceCredentials)(this._credentials, this._datasource.type))
            .then(({ Format, history }) => {
            if (Format !== this.vault.format.getFormat()) {
                throw new Error("Loaded format does not match that of current vault");
            }
            return Format.historiesDiffer(this.vault.format.history, history);
        });
    }
    /**
     * Lock the source
     * @memberof VaultSource
     */
    lock() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.status !== VaultSource.STATUS_UNLOCKED) {
                throw new layerr_1.Layerr(`Failed locking source: Source in invalid state (${this.status}): ${this.id}`);
            }
            yield this._enqueueStateChange(() => __awaiter(this, void 0, void 0, function* () {
                this._status = VaultSource.STATUS_PENDING;
                const currentCredentials = this._credentials;
                const currentVault = this._vault;
                const currentDatasource = this._datasource;
                const currentAttachmentMgr = this._attachmentManager;
                try {
                    const credentialsStr = yield this._credentials.toSecureString();
                    this._credentials = credentialsStr;
                    this._datasource = null;
                    this._vault = null;
                    this._attachmentManager = null;
                    this._status = VaultSource.STATUS_LOCKED;
                    this.emit("locked");
                }
                catch (err) {
                    this._credentials = currentCredentials;
                    this._datasource = currentDatasource;
                    this._vault = currentVault;
                    this._status = VaultSource.STATUS_UNLOCKED;
                    this._attachmentManager = currentAttachmentMgr;
                    throw new layerr_1.Layerr(err, "Failed locking source");
                }
            }));
        });
    }
    /**
     * Merge remote contents
     * Detects differences between a local and a remote item, and merges the
     * two copies together.
     * @returns A promise that resolves with the newly merged archive -
     *      This archive is automatically saved over the original local copy.
     * @memberof VaultSource
     */
    mergeFromRemote() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this._datasource.type !== "text") {
                // Only clear if not a TextDatasource
                this._datasource.setContent("");
            }
            const { Format, history } = yield this._datasource.load((0, register_1.prepareDatasourceCredentials)(this._credentials, this._datasource.type));
            if (Format !== this._vault.format.getFormat()) {
                throw new Error("Format loaded during merge did not match current");
            }
            const newVault = Format.vaultFromMergedHistories(history, this._vault.format.history);
            this._vault._updateFormat(newVault.format);
            return this._vault;
        });
    }
    /**
     * Rename the vault source
     * @param name The new name
     * @memberof VaultSource
     */
    rename(name) {
        this._name = name;
        this.emit("updated");
    }
    /**
     * Save the vault to the remote, ensuring that it's first merged and
     * updated to prevent conflicts or overwrites.
     * @memberof VaultSource
     */
    save(config = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            const { storeOfflineCopy = true } = config;
            yield this._enqueueStateChange(() => __awaiter(this, void 0, void 0, function* () {
                if (yield this.localDiffersFromRemote()) {
                    yield this.mergeFromRemote();
                }
                // Capture encrypted content
                let encryptedContent = null;
                const encryptedCallback = ({ content }) => {
                    encryptedContent = content;
                };
                this._datasource.once("encryptedContent", encryptedCallback);
                // Unlock
                try {
                    yield this._datasource.save(this._vault.format.history, (0, register_1.prepareDatasourceCredentials)(this._credentials, this._datasource.type));
                }
                catch (err) {
                    this._datasource.off("encryptedContent", encryptedCallback);
                    throw err;
                }
                // Clear state
                this._vault.format.dirty = false;
                // Handle offline state
                if (storeOfflineCopy && encryptedContent) {
                    // Store an offline copy for later use
                    yield (0, vaultManagement_1.storeSourceOfflineCopy)(this._vaultManager._cacheStorage, this.id, encryptedContent);
                }
                // Misc
                yield this._updateInsights();
            }), /* stack */ "saving");
            this.emit("updated");
        });
    }
    supportsAttachments() {
        if (this.status !== VaultSource.STATUS_UNLOCKED)
            return false;
        return this._datasource.supportsAttachments();
    }
    testMasterPassword(password) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.status !== types_1.VaultSourceStatus.Locked && this.status !== types_1.VaultSourceStatus.Unlocked) {
                throw new Error(`Source in invalid state for password test: ${this.status}`);
            }
            const credStr = this.status === types_1.VaultSourceStatus.Locked
                ? this._credentials
                : yield this._credentials.toSecureString();
            try {
                yield processDehydratedCredentials(credStr, password);
                return true;
            }
            catch (err) {
                return false;
            }
        });
    }
    unlock(vaultCredentials, config = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!Credentials_1.default.isCredentials(vaultCredentials)) {
                throw new layerr_1.Layerr(`Failed unlocking source: Invalid credentials passed to source: ${this.id}`);
            }
            const { initialiseRemote = false, loadOfflineCopy = false, storeOfflineCopy = true } = config;
            if (this.status !== VaultSource.STATUS_LOCKED) {
                throw new layerr_1.Layerr(`Failed unlocking source: Source in invalid state (${this.status}): ${this.id}`);
            }
            const { masterPassword } = (0, channel_1.getCredentials)(vaultCredentials.id);
            const originalCredentials = this._credentials;
            this._status = VaultSource.STATUS_PENDING;
            yield this._enqueueStateChange(() => __awaiter(this, void 0, void 0, function* () {
                // Get offline content if available and requested
                const offlineContent = loadOfflineCopy ? yield this.getOfflineContent() : null;
                const credentials = (this._credentials = yield processDehydratedCredentials(this._credentials, masterPassword));
                // Initialise datasource
                const datasource = (this._datasource = (0, register_1.credentialsToDatasource)(Credentials_1.default.fromCredentials(credentials, masterPassword)));
                datasource.sourceID = this.id;
                if (typeof offlineContent === "string") {
                    datasource.setContent(offlineContent);
                }
                // Listen for datasource updates
                datasource.on("updated", () => {
                    this._waitNonPending()
                        .then(() => __awaiter(this, void 0, void 0, function* () {
                        if (this.status === VaultSource.STATUS_UNLOCKED) {
                            yield this._updateCredentialsFromDatasource();
                        }
                        this.emit("updated");
                    }))
                        .catch(err => {
                        console.error(`Error updating datasource credentials for vault: ${this.id}`, err);
                    });
                });
                // Perform pre-save or load
                if (initialiseRemote) {
                    const defaultVault = Vault_1.default.createWithDefaults();
                    yield datasource.save(defaultVault.format.history, credentials);
                    this._vault = defaultVault;
                }
                else {
                    const { Format, history } = yield datasource.load(credentials);
                    this._vault = Vault_1.default.createFromHistory(history, Format);
                }
                // Handle offline state
                if (storeOfflineCopy) {
                    // Store an offline copy for later use
                    yield (0, vaultManagement_1.storeSourceOfflineCopy)(this._vaultManager._cacheStorage, this.id, datasource._content);
                }
                if (loadOfflineCopy) {
                    // Flag the format as read-only
                    this.vault.format._readOnly = true;
                }
                // Configure source status
                this._status = VaultSource.STATUS_UNLOCKED;
                this._attachmentManager = new AttachmentManager_1.default(this);
                this.emit("unlocked");
            })).catch(err => {
                this._status = VaultSource.STATUS_LOCKED;
                this._vault = null;
                this._datasource = null;
                this._credentials = originalCredentials;
                this._attachmentManager = null;
                throw new layerr_1.Layerr(err, "Failed unlocking source");
            });
        });
    }
    /**
     * Update the vault
     * @returns A promise that resolves once the update has
     *  completed
     * @memberof VaultSource
     */
    update({ skipDiff = false } = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            const didUpdate = yield this._enqueueStateChange(() => (skipDiff ? Promise.resolve(false) : this.localDiffersFromRemote()).then(differs => {
                if (differs) {
                    return this.mergeFromRemote().then(() => true);
                }
                return false;
            }), 
            // @todo shares
            // .then(() => initialiseShares(this)),
            /* stack */ "updating");
            if (didUpdate) {
                this.emit("updated");
            }
        });
    }
    /**
     * Write the vault to the remote
     * - This does not perform any merging or sync checks, but simply
     * writes the vault contents to the remote, overwriting whatever
     * was there before.
     * @returns A promise that resolves when saving has completed
     * @memberof VaultSource
     */
    write() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this._enqueueStateChange(() => __awaiter(this, void 0, void 0, function* () {
                yield this._datasource.save(this._vault.format.history, (0, register_1.prepareDatasourceCredentials)(this._credentials, this._datasource.type));
                this._vault.format.dirty = false;
                yield this._updateInsights();
            }), /* stack */ "saving");
            this.emit("updated");
        });
    }
    _applyShares() {
        // @todo
        // this._shares.forEach(share => {
        //     if (!share.archiveHasAppliedShare(this.archive)) {
        //         share.applyToArchive(this.archive);
        //     }
        // });
    }
    _enqueueStateChange(cb, stack) {
        const channel = this._queue.channel("state");
        return stack ? channel.enqueue(cb, undefined, stack) : channel.enqueue(cb);
    }
    _unloadShares() {
        const Format = this.vault.format.getFormat();
        const extractedShares = Format.extractSharesFromHistory(this.vault.format.history);
        // Reset archive history (without shares)
        const { base } = extractedShares;
        delete extractedShares.base;
        this.vault.format.erase();
        this.vault.format.execute(base);
        // Update share payloads
        Object.keys(extractedShares).forEach(shareID => {
            const share = this._shares.find(share => share.id === shareID);
            if (!share) {
                throw new Error(`Failed updating extracted share: No share found in workspace for ID: ${shareID}`);
            }
            share.updateHistory(extractedShares[shareID]);
        });
    }
    _updateCredentialsFromDatasource() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.status !== VaultSource.STATUS_UNLOCKED) {
                throw new layerr_1.Layerr(`Failed updating source credentials: Source is not unlocked: ${this.id}`);
            }
            const { masterPassword } = (0, channel_1.getCredentials)(this._credentials.id);
            this._credentials = Credentials_1.default.fromCredentials(this._datasource.credentials, masterPassword);
        });
    }
    _updateInsights() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.status !== VaultSource.STATUS_UNLOCKED) {
                throw new layerr_1.Layerr(`Failed updating vault insights: Source is not unlocked: ${this.id}`);
            }
            const insights = (0, vault_1.generateVaultInsights)(this.vault);
            yield this._datasource.updateInsights(insights);
        });
    }
    _updateVaultCredentials(newCredentials) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.status !== VaultSource.STATUS_UNLOCKED) {
                throw new layerr_1.Layerr(`Failed updating vault credentials: Source is not unlocked: ${this.id}`);
            }
            this._credentials = newCredentials;
            yield this.write();
        });
    }
    _waitNonPending() {
        return new Promise(resolve => {
            if (this.status !== VaultSource.STATUS_PENDING)
                return resolve();
            const handleChange = () => {
                this.removeListener("unlocked", handleChange);
                this.removeListener("locked", handleChange);
                resolve();
            };
            this.on("unlocked", handleChange);
            this.on("locked", handleChange);
        });
    }
}
exports.default = VaultSource;
VaultSource.STATUS_LOCKED = types_1.VaultSourceStatus.Locked;
VaultSource.STATUS_PENDING = types_1.VaultSourceStatus.Pending;
VaultSource.STATUS_UNLOCKED = types_1.VaultSourceStatus.Unlocked;
