"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const uuid_1 = require("../tools/uuid");
const channel_1 = require("./channel");
const appEnv_1 = require("../env/appEnv");
/**
 * The signature of legacy encrypted credentials
 * @private
 */
const LEGACY_SIGNING_KEY = "b~>buttercup/acreds.v2.";
/**
 * The signature of encrypted credentials
 * @private
 */
const SIGNING_KEY = "bc~3>";
/**
 * Sign encrypted content
 * @see SIGNING_KEY
 * @private
 * @param content The encrypted text
 * @returns The signed key
 */
function signEncryptedContent(content) {
    return `${SIGNING_KEY}${content}`;
}
/**
 * Remove the signature from encrypted content
 * @private
 * @param content The encrypted text
 * @returns The unsigned encrypted key
 * @throws {Error} Throws if no SIGNING_KEY is detected
 * @see SIGNING_KEY
 */
function unsignEncryptedContent(content) {
    const newIndex = content.indexOf(SIGNING_KEY);
    const oldIndex = content.indexOf(LEGACY_SIGNING_KEY);
    if (newIndex === -1 && oldIndex === -1) {
        throw new Error("Invalid credentials content (unknown signature)");
    }
    return newIndex === 0 ? content.substr(SIGNING_KEY.length) : content.substr(LEGACY_SIGNING_KEY.length);
}
/**
 * Secure credentials storage/transfer class
 * - Allows for the safe transfer of credentials within the
 * Buttercup application environment. Will not allow
 * credentials to be shared or transferred outside of the
 * environment. Credential properties are stored in memory
 * and are inaccessible to public functions.
 */
class Credentials {
    /**
     * Get all available purposes
     * @static
     */
    static allPurposes() {
        return [
            Credentials.PURPOSE_DECRYPT_VAULT,
            Credentials.PURPOSE_ENCRYPT_VAULT,
            Credentials.PURPOSE_SECURE_EXPORT
        ];
    }
    /**
     * Create a new Credentials instance using an existing Credentials
     * instance - This can be used to reset a credentials's purposes.
     * @param credentials A credentials instance
     * @param masterPassword The master password used to
     *  encrypt the instance being cloned
     * @throws {Error} Throws if no master password provided
     * @throws {Error} Throws if master password does not match
     *  original
     */
    static fromCredentials(credentials, masterPassword) {
        if (!masterPassword) {
            throw new Error("Master password is required for credentials cloning");
        }
        const credentialsData = (0, channel_1.getCredentials)(credentials.id);
        if (credentialsData.masterPassword !== masterPassword) {
            throw new Error("Master password does not match that of the credentials to be cloned");
        }
        const newData = JSON.parse(JSON.stringify(credentialsData.data));
        return new Credentials(newData, masterPassword);
    }
    /**
     * Create a new Credentials instance from a Datasource configuration
     * @param datasourceConfig The configuration for the
     *  datasource - this usually includes the credential data used for
     *  authenticating against the datasource host platform.
     * @param masterPassword Optional master password to
     *  store alongside the credentials. Used to create secure strings.
     */
    static fromDatasource(datasourceConfig, masterPassword = null) {
        return new Credentials({
            datasource: datasourceConfig
        }, masterPassword);
    }
    /**
     * Create a new Credentials instance from a password
     * - uses the single password value as the master password stored
     * alongside the original password if no master password is
     * provided. The master password is used when generating secure
     * strings.
     * @param password The password to store
     * @param masterPassword Optional master password
     *  to store alongside the credentials. Used to create secure
     *  strings.
     */
    static fromPassword(password, masterPassword = null) {
        const masterPass = masterPassword || password;
        return new Credentials({ password }, masterPass);
    }
    /**
     * Create a new instance from a secure string
     * @param content Encrypted content
     * @param masterPassword The password for decryption
     * @returns A promise that resolves with the new instance
     */
    static fromSecureString(content, masterPassword) {
        const decrypt = (0, appEnv_1.getSharedAppEnv)().getProperty("crypto/v1/decryptText");
        return decrypt(unsignEncryptedContent(content), masterPassword)
            .then((decryptedContent) => JSON.parse(decryptedContent))
            .then((credentialsData) => {
            // Handle compatibility updates for legacy credentials
            if (credentialsData.datasource) {
                if (typeof credentialsData.datasource === "string") {
                    credentialsData.datasource = JSON.parse(credentialsData.datasource);
                }
                // Move username and password INTO the datasource config, as
                // they relate to the remote connection/source
                if (credentialsData.username) {
                    credentialsData.datasource.username = credentialsData.username;
                    delete credentialsData.username;
                }
                if (credentialsData.password) {
                    credentialsData.datasource.password = credentialsData.password;
                    delete credentialsData.password;
                }
            }
            return new Credentials(credentialsData, masterPassword);
        });
    }
    /**
     * Check if a value is an instance of Credentials
     * @param inst The value to check
     */
    static isCredentials(inst) {
        return !!inst && typeof inst === "object" && typeof inst.toSecureString === "function" && !!inst.id;
    }
    /**
     * Create a new Credentials instance
     * @param obj Object data representing some credentials
     * @param masterPassword Optional master password to store with
     *  the credentials data, which is used for generating secure strings.
     */
    constructor(obj = {}, masterPassword = null) {
        const id = (0, uuid_1.generateUUID)();
        Object.defineProperty(this, "id", {
            writable: false,
            configurable: false,
            enumerable: true,
            value: id
        });
        (0, channel_1.setCredentials)(id, {
            data: obj,
            masterPassword,
            purposes: Credentials.allPurposes(),
            open: false
        });
    }
    /**
     * Get raw credentials data (only available in specialised environments)
     * @returns Credentials data object, or null if not available
     */
    getCredentialsData() {
        const isClosedEnv = (0, appEnv_1.getSharedAppEnv)().getProperty("env/v1/isClosedEnv")();
        const payload = (0, channel_1.getCredentials)(this.id);
        if (isClosedEnv || payload.open === true) {
            return payload.data;
        }
        return null;
    }
    /**
     * Get raw credentials data (only available in specialised environments)
     * @deprecated Use getCredentialsData instead
     */
    getData() {
        const isClosedEnv = (0, appEnv_1.getSharedAppEnv)().getProperty("env/v1/isClosedEnv")();
        const payload = (0, channel_1.getCredentials)(this.id);
        if (isClosedEnv || payload.open === true) {
            return payload;
        }
        return null;
    }
    /**
     * Restrict the purposes that this set of credentials
     * can be used for. Once a purpose is removed it can
     * no longer be added again to the same instance.
     * @param allowedPurposes An array of
     *  new allowed purposes. If a purpose mentioned is
     *  not currently permitted, it will be ignored.
     * @returns Returns self
     * @example
     *  credentials.restrictPurposes(
     *      Credentials.PURPOSE_SECURE_EXPORT
     *  );
     *  // credentials can only be exported to an
     *  // encrypted string, and not used for things
     *  // like encrypting datasource changes.
     */
    restrictPurposes(allowedPurposes) {
        const creds = (0, channel_1.getCredentials)(this.id);
        const { purposes } = creds;
        // Filter out purposes which have already been restricted
        const finalPurposes = allowedPurposes.filter(newPurpose => purposes.includes(newPurpose));
        (0, channel_1.setCredentials)(this.id, Object.assign(creds, {
            purposes: finalPurposes
        }));
        return this;
    }
    /**
     * Set the credential data
     * @param data The credentials data to overwrite the old with
     */
    setCredentialsData(data) {
        const isClosedEnv = (0, appEnv_1.getSharedAppEnv)().getProperty("env/v1/isClosedEnv")();
        const payload = (0, channel_1.getCredentials)(this.id);
        if (!isClosedEnv && payload.open !== true) {
            throw new Error("Unable to set data: Insecure environment and payload is not open");
        }
        const newPayload = Object.assign(Object.assign({}, payload), { data });
        (0, channel_1.setCredentials)(this.id, newPayload);
    }
    /**
     * Convert the credentials to an encrypted string, for storage
     * @returns A promise that resolves with the encrypted credentials
     * @throws {Error} Rejects when masterPassword is not a string
     * @throws {Error} Rejects if credentials don't permit secure export purposes
     */
    toSecureString() {
        return __awaiter(this, void 0, void 0, function* () {
            if ((0, channel_1.credentialsAllowsPurpose)(this.id, Credentials.PURPOSE_SECURE_EXPORT) !== true) {
                throw new Error("Credential purposes don't allow for secure exports");
            }
            const encrypt = (0, appEnv_1.getSharedAppEnv)().getProperty("crypto/v1/encryptText");
            const { data, masterPassword } = (0, channel_1.getCredentials)(this.id);
            if (typeof masterPassword !== "string") {
                throw new Error("Cannot convert Credentials to string: master password was not set or is invalid");
            }
            return encrypt(JSON.stringify(data), masterPassword).then(signEncryptedContent);
        });
    }
    /**
     * Get raw credentials data (only available in specialised environments)
     * @protected
     * @deprecated
     */
    _getData() {
        return this.getData();
    }
}
exports.default = Credentials;
Credentials.PURPOSE_DECRYPT_VAULT = "vault-decrypt";
Credentials.PURPOSE_ENCRYPT_VAULT = "vault-encrypt";
Credentials.PURPOSE_SECURE_EXPORT = "secure-export";
