"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const TextDatasource_1 = __importDefault(require("./TextDatasource"));
const register_1 = require("./register");
const channel_1 = require("../credentials/channel");
const appEnv_1 = require("../env/appEnv");
/**
 * Datasource for Dropbox archives
 * @augments TextDatasource
 * @memberof module:Buttercup
 */
class DropboxDatasource extends TextDatasource_1.default {
    /**
     * Datasource for Dropbox accounts
     * @param credentials Credentials instance to configure the
     *  datsource with
     */
    constructor(credentials) {
        super(credentials);
        const { data: credentialData } = (0, channel_1.getCredentials)(credentials.id);
        const { datasource: datasourceConfig } = credentialData;
        const { token, path } = datasourceConfig;
        this.path = path;
        this.token = token;
        const createClient = (0, appEnv_1.getSharedAppEnv)().getProperty("net/dropbox/v1/newClient");
        this.client = createClient(token);
        this.type = "dropbox";
        (0, register_1.fireInstantiationHandlers)("dropbox", this);
    }
    /**
     * Get the datasource configuration
     * @memberof DropboxDatasource
     */
    getConfiguration() {
        return {
            type: "dropbox",
            token: this.token,
            path: this.path
        };
    }
    /**
     * Load an archive from the datasource
     * @param credentials The credentials for decryption
     * @returns A promise that resolves archive history
     * @memberof DropboxDatasource
     */
    load(credentials) {
        if (this.hasContent) {
            return super.load(credentials);
        }
        return this.client.getFileContents(this.path).then(content => {
            this.setContent(content);
            return super.load(credentials);
        });
    }
    /**
     * Save an archive using the datasource
     * @param history The archive history to save
     * @param credentials The credentials to save with
     * @returns A promise that resolves when saving has completed
     * @memberof DropboxDatasource
     */
    save(history, credentials) {
        return super
            .save(history, credentials)
            .then((encryptedContent) => this.client.putFileContents(this.path, encryptedContent));
    }
    /**
     * Whether or not the datasource supports bypassing remote fetch operations
     * @returns True if content can be set to bypass fetch operations,
     *  false otherwise
     * @memberof DropboxDatasource
     */
    supportsRemoteBypass() {
        return true;
    }
}
exports.default = DropboxDatasource;
(0, register_1.registerDatasource)("dropbox", DropboxDatasource);
