"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const pify_1 = __importDefault(require("pify"));
const TextDatasource_1 = __importDefault(require("./TextDatasource"));
const register_1 = require("./register");
const channel_1 = require("../credentials/channel");
const attachments_1 = require("../tools/attachments");
/**
 * File datasource for loading and saving files
 * @augments TextDatasource
 * @memberof module:Buttercup
 */
class FileDatasource extends TextDatasource_1.default {
    /**
     * Constructor for the file datasource
     * @param credentials The credentials instance with which to
     *  use to configure the datasource
     */
    constructor(credentials) {
        super(credentials);
        const { data: credentialData } = (0, channel_1.getCredentials)(credentials.id);
        const { datasource: datasourceConfig } = credentialData;
        const { path } = datasourceConfig;
        this._filename = path;
        this.mkdir = (0, pify_1.default)(fs_1.default.mkdir);
        this.readFile = (0, pify_1.default)(fs_1.default.readFile);
        this.stat = (0, pify_1.default)(fs_1.default.stat);
        this.unlink = (0, pify_1.default)(fs_1.default.unlink);
        this.writeFile = (0, pify_1.default)(fs_1.default.writeFile);
        this.type = "file";
        (0, register_1.fireInstantiationHandlers)("file", this);
    }
    get baseDir() {
        return path_1.default.dirname(this.path);
    }
    /**
     * The file path
     * @memberof FileDatasource
     */
    get path() {
        return this._filename;
    }
    /**
     * Ensure attachment paths exist
     * @memberof FileDatasource
     * @protected
     */
    _ensureAttachmentsPaths(vaultID) {
        return __awaiter(this, void 0, void 0, function* () {
            const attachmentsDir = path_1.default.join(this.baseDir, ".buttercup", vaultID);
            yield this.mkdir(attachmentsDir, { recursive: true });
        });
    }
    /**
     * Get encrypted attachment
     * - Loads the attachment contents from a file into a buffer
     * @param vaultID The ID of the vault
     * @param attachmentID The ID of the attachment
     * @memberof FileDatasource
     */
    getAttachment(vaultID, attachmentID) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this._ensureAttachmentsPaths(vaultID);
            const attachmentPath = path_1.default.join(this.baseDir, ".buttercup", vaultID, `${attachmentID}.${attachments_1.ATTACHMENT_EXT}`);
            return this.readFile(attachmentPath);
        });
    }
    /**
     * Get the datasource configuration
     * @memberof FileDatasource
     */
    getConfiguration() {
        return {
            type: "file",
            path: this._filename
        };
    }
    /**
     * Load from the filename specified in the constructor using a password
     * @param credentials The credentials for decryption
     * @returns A promise resolving with archive history
     * @memberof FileDatasource
     */
    load(credentials) {
        return this.hasContent
            ? super.load(credentials)
            : this.readFile(this.path, "utf8").then(contents => {
                this.setContent(contents);
                return super.load(credentials);
            });
    }
    /**
     * Put attachment data
     * @param vaultID The ID of the vault
     * @param attachmentID The ID of the attachment
     * @param buffer The attachment data
     * @param details The attachment details
     * @memberof FileDatasource
     */
    putAttachment(vaultID, attachmentID, buffer, details) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this._ensureAttachmentsPaths(vaultID);
            const attachmentPath = path_1.default.join(this.baseDir, ".buttercup", vaultID, `${attachmentID}.${attachments_1.ATTACHMENT_EXT}`);
            yield this.writeFile(attachmentPath, buffer);
        });
    }
    /**
     * Remove an attachment
     * @param vaultID The ID of the vault
     * @param attachmentID The ID of the attachment
     * @memberof FileDatasource
     */
    removeAttachment(vaultID, attachmentID) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this._ensureAttachmentsPaths(vaultID);
            const attachmentPath = path_1.default.join(this.baseDir, ".buttercup", vaultID, `${attachmentID}.${attachments_1.ATTACHMENT_EXT}`);
            yield this.unlink(attachmentPath);
        });
    }
    /**
     * Save archive history to a file
     * @param history The archive history to save
     * @param credentials The credentials to save with
     * @returns A promise that resolves when saving is complete
     * @memberof FileDatasource
     */
    save(history, credentials) {
        return super.save(history, credentials).then(encrypted => this.writeFile(this.path, encrypted));
    }
    /**
     * Whether or not the datasource supports attachments
     * @memberof FileDatasource
     */
    supportsAttachments() {
        return true;
    }
    /**
     * Whether or not the datasource supports bypassing remote fetch operations
     * @returns True if content can be set to bypass fetch operations,
     *  false otherwise
     * @memberof FileDatasource
     */
    supportsRemoteBypass() {
        return true;
    }
}
exports.default = FileDatasource;
(0, register_1.registerDatasource)("file", FileDatasource);
