"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const googledrive_client_1 = require("@buttercup/googledrive-client");
const layerr_1 = require("layerr");
const DatasourceAuthManager_1 = __importDefault(require("./DatasourceAuthManager"));
const TextDatasource_1 = __importDefault(require("./TextDatasource"));
const register_1 = require("./register");
const channel_1 = require("../credentials/channel");
const DATASOURCE_TYPE = "googledrive";
/**
 * Datasource for Google Drive archives
 * @augments TextDatasource
 * @memberof module:Buttercup
 */
class GoogleDriveDatasource extends TextDatasource_1.default {
    /**
     * Datasource for Google Drive connections
     * @param {Credentials} credentials The credentials instance with which to
     *  configure the datasource with
     */
    constructor(credentials) {
        super(credentials);
        const { data: credentialData } = (0, channel_1.getCredentials)(credentials.id);
        const { datasource: datasourceConfig } = credentialData;
        const { token, refreshToken, fileID } = datasourceConfig;
        this.fileID = fileID;
        this.token = token;
        this.refreshToken = refreshToken;
        this.client = (0, googledrive_client_1.createClient)(token);
        this.authManager = DatasourceAuthManager_1.default.getSharedManager();
        this.type = DATASOURCE_TYPE;
        (0, register_1.fireInstantiationHandlers)(DATASOURCE_TYPE, this);
    }
    /**
     * Get the datasource configuration
     * @memberof GoogleDriveDatasource
     */
    getConfiguration() {
        return {
            type: "googledrive",
            token: this.token,
            refreshToken: this.refreshToken,
            fileID: this.fileID
        };
    }
    /**
     * Load an archive from the datasource
     * @param {Credentials} credentials The credentials for decryption
     * @returns {Promise.<LoadedVaultData>} A promise that resolves archive history
     * @memberof GoogleDriveDatasource
     */
    load(credentials, hasAuthed = false) {
        if (this.hasContent) {
            return super.load(credentials);
        }
        return this.client
            .getFileContents(this.fileID)
            .then(content => {
            this.setContent(content);
            return super.load(credentials);
        })
            .catch(err => {
            const { authFailure = false } = layerr_1.Layerr.info(err);
            if (!authFailure) {
                throw new layerr_1.Layerr(err, "Failed fetching Google Drive vault");
            }
            else if (hasAuthed) {
                throw new layerr_1.Layerr(err, "Re-authentication failed");
            }
            return this.authManager
                .executeAuthHandlers(DATASOURCE_TYPE, this)
                .then(() => this.load(credentials, true))
                .catch(err2 => {
                throw new layerr_1.Layerr(err2, "Failed fetching Google Drive vault");
            });
        });
    }
    /**
     * Save an archive using the datasource
     * @param {Array.<String>} history The archive history to save
     * @param {Credentials} credentials The credentials to save with
     * @returns {Promise} A promise that resolves when saving has completed
     * @memberof GoogleDriveDatasource
     */
    save(history, credentials, hasAuthed = false) {
        return super
            .save(history, credentials)
            .then(encryptedContent => this.client.putFileContents({
            id: this.fileID,
            contents: encryptedContent
        }))
            .catch(err => {
            const { authFailure = false } = layerr_1.Layerr.info(err);
            if (!authFailure) {
                throw new layerr_1.Layerr(err, "Failed saving Google Drive vault");
            }
            else if (hasAuthed) {
                throw new layerr_1.Layerr(err, "Re-authentication failed");
            }
            return this.authManager
                .executeAuthHandlers(DATASOURCE_TYPE, this)
                .then(() => this.save(history, credentials, true))
                .catch(err2 => {
                throw new layerr_1.Layerr(err2, "Failed saving Google Drive vault");
            });
        });
    }
    /**
     * Whether or not the datasource supports bypassing remote fetch operations
     * @returns {Boolean} True if content can be set to bypass fetch operations,
     *  false otherwise
     * @memberof GoogleDriveDatasource
     */
    supportsRemoteBypass() {
        return true;
    }
    /**
     * Update the OAuth2 tokens
     * @param accessToken The access token
     * @param refreshToken The refresh token, if available
     * @memberof GoogleDriveDatasource
     */
    updateTokens(accessToken, refreshToken) {
        this.token = accessToken;
        if (refreshToken)
            this.refreshToken = refreshToken;
        this.client = (0, googledrive_client_1.createClient)(accessToken);
        const { data: credentialData } = (0, channel_1.getCredentials)(this.credentials.id);
        credentialData.datasource.token = accessToken;
        if (refreshToken)
            credentialData.datasource.refreshToken = refreshToken;
        this.emit("updated");
    }
}
exports.default = GoogleDriveDatasource;
(0, register_1.registerDatasource)(DATASOURCE_TYPE, GoogleDriveDatasource);
