"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_posix_1 = __importDefault(require("path-posix"));
const TextDatasource_1 = __importDefault(require("./TextDatasource"));
const register_1 = require("./register");
const appEnv_1 = require("../env/appEnv");
const channel_1 = require("../credentials/channel");
const attachments_1 = require("../tools/attachments");
const MAX_DATA_SIZE = 200 * 1024 * 1024; // 200 MB
/**
 * WebDAV datasource for reading and writing remote archives
 * @augments TextDatasource
 * @memberof module:Buttercup
 */
class WebDAVDatasource extends TextDatasource_1.default {
    /**
     * Constructor for the datasource
     * @param credentials Credentials for the datasource
     */
    constructor(credentials) {
        super(credentials);
        const { data: credentialData } = (0, channel_1.getCredentials)(credentials.id);
        const { datasource: datasourceConfig } = credentialData;
        const { endpoint, password, path, username } = (this._config = datasourceConfig);
        this._path = path;
        const createClient = (0, appEnv_1.getSharedAppEnv)().getProperty("net/webdav/v1/newClient");
        if (typeof username === "string" && typeof password === "string") {
            this._client = createClient(endpoint, {
                username,
                password,
                maxBodyLength: MAX_DATA_SIZE,
                maxContentLength: MAX_DATA_SIZE
            });
        }
        else {
            this._client = createClient(endpoint, {
                maxBodyLength: MAX_DATA_SIZE,
                maxContentLength: MAX_DATA_SIZE
            });
        }
        this.type = "webdav";
        (0, register_1.fireInstantiationHandlers)("webdav", this);
    }
    /**
     * The vault file's base directory
     * @memberof WebDAVDatasource
     */
    get baseDir() {
        return path_posix_1.default.dirname(this.path);
    }
    /**
     * The WebDAV client instance
     * @memberof WebDAVDatasource
     */
    get client() {
        return this._client;
    }
    /**
     * The remote archive path
     * @memberof WebDAVDatasource
     */
    get path() {
        return this._path;
    }
    /**
     * Ensure attachment paths exist
     * @memberof WebDAVDatasource
     * @protected
     */
    _ensureAttachmentsPaths(vaultID) {
        return __awaiter(this, void 0, void 0, function* () {
            const attachmentsDir = path_posix_1.default.join(this.baseDir, ".buttercup", vaultID);
            yield this.client.createDirectory(attachmentsDir, { recursive: true });
        });
    }
    /**
     * Get encrypted attachment
     * - Loads the attachment contents from a file into a buffer
     * @param vaultID The ID of the vault
     * @param attachmentID The ID of the attachment
     * @memberof WebDAVDatasource
     */
    getAttachment(vaultID, attachmentID) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this._ensureAttachmentsPaths(vaultID);
            const attachmentPath = path_posix_1.default.join(this.baseDir, ".buttercup", vaultID, `${attachmentID}.${attachments_1.ATTACHMENT_EXT}`);
            return this.client.getFileContents(attachmentPath);
        });
    }
    /**
     * Get the datasource configuration
     * @memberof WebDAVDatasource
     */
    getConfiguration() {
        return this._config;
    }
    /**
     * Load archive history from the datasource
     * @param credentials The credentials for archive decryption
     * @returns A promise resolving archive history
     * @memberof WebDAVDatasource
     */
    load(credentials) {
        return this.hasContent
            ? super.load(credentials)
            : this.client.getFileContents(this.path, { format: "text" }).then(content => {
                this.setContent(content);
                return super.load(credentials);
            });
    }
    /**
     * Put attachment data
     * @param vaultID The ID of the vault
     * @param attachmentID The ID of the attachment
     * @param buffer The attachment data
     * @param details The attachment details
     * @memberof WebDAVDatasource
     */
    putAttachment(vaultID, attachmentID, buffer, details) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this._ensureAttachmentsPaths(vaultID);
            const attachmentPath = path_posix_1.default.join(this.baseDir, ".buttercup", vaultID, `${attachmentID}.${attachments_1.ATTACHMENT_EXT}`);
            yield this.client.putFileContents(attachmentPath, buffer);
        });
    }
    /**
     * Remove an attachment
     * @param vaultID The ID of the vault
     * @param attachmentID The ID of the attachment
     * @memberof WebDAVDatasource
     */
    removeAttachment(vaultID, attachmentID) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this._ensureAttachmentsPaths(vaultID);
            const attachmentPath = path_posix_1.default.join(this.baseDir, ".buttercup", vaultID, `${attachmentID}.${attachments_1.ATTACHMENT_EXT}`);
            yield this.client.deleteFile(attachmentPath);
        });
    }
    /**
     * Save archive contents to the WebDAV service
     * @param history Archive history
     * @param credentials The credentials for encryption
     * @returns A promise resolving when the save is complete
     * @memberof WebDAVDatasource
     */
    save(history, credentials) {
        const _super = Object.create(null, {
            save: { get: () => super.save }
        });
        return __awaiter(this, void 0, void 0, function* () {
            const content = yield _super.save.call(this, history, credentials);
            yield this.client.putFileContents(this.path, content);
        });
    }
    /**
     * Whether or not the datasource supports attachments
     * @memberof WebDAVDatasource
     */
    supportsAttachments() {
        return true;
    }
    /**
     * Whether or not the datasource supports bypassing remote fetch operations
     * @returns True if content can be set to bypass fetch operations,
     *  false otherwise
     * @memberof WebDAVDatasource
     */
    supportsRemoteBypass() {
        return true;
    }
}
exports.default = WebDAVDatasource;
(0, register_1.registerDatasource)("webdav", WebDAVDatasource);
