"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.setEntryPropertyValueType = exports.idSignifiesNew = exports.getEntryPropertyValueType = exports.getEntryValue = exports.createFieldDescriptor = void 0;
const Entry_1 = __importDefault(require("../core/Entry"));
const uuid_1 = require("../tools/uuid");
const types_1 = require("../types");
const VALID_VALUE_TYPES = [
    types_1.EntryPropertyValueType.Note,
    types_1.EntryPropertyValueType.OTP,
    types_1.EntryPropertyValueType.Password,
    types_1.EntryPropertyValueType.Text
];
/**
 * Create a descriptor for a field to be used within a facade
 * @param entry The entry instance to process or null if the initial value
 *  should be empty
 * @param title The field title
 * @param entryPropertyType The type of entry property (property/attribute)
 * @param entryPropertyName The name of the property
 * @param options The options for the field
 * @returns The field descriptor
 * @memberof module:Buttercup
 */
function createFieldDescriptor(entry, title, entryPropertyType, entryPropertyName, options = {}) {
    const { formatting = false, removeable = false, valueType = null } = options;
    const value = entry ? getEntryValue(entry, entryPropertyType, entryPropertyName) : "";
    // Return descriptor
    return {
        id: (0, uuid_1.generateUUID)(),
        title,
        propertyType: entryPropertyType,
        property: entryPropertyName,
        value,
        valueType: valueType
            ? valueType
            : entryPropertyType === "attribute"
                ? null
                : getEntryPropertyValueType(entry, entryPropertyName),
        formatting,
        removeable
    };
}
exports.createFieldDescriptor = createFieldDescriptor;
/**
 * Get a value on an entry for a specific property type
 * @param entry The entry instance
 * @param propertyType The type of entry property (property/attribute)
 * @param name The property name
 * @returns The property value
 * @throws {Error} Throws for unknown property types
 * @deprecated Not in use - To be removed
 */
function getEntryValue(entry, propertyType, name) {
    switch (propertyType) {
        case types_1.EntryPropertyType.Property:
            return entry.getProperty(name);
        case types_1.EntryPropertyType.Attribute:
            return entry.getAttribute(name);
        default:
            throw new Error(`Cannot retrieve value: Unknown property type: ${propertyType}`);
    }
}
exports.getEntryValue = getEntryValue;
/**
 * Get the entry value type
 * @param entry Entry instance
 * @param propertyName The entry property name
 * @returns The entry value type (returns default "text"
 *  if entry not specified)
 */
function getEntryPropertyValueType(entry, propertyName) {
    if (!entry) {
        return types_1.EntryPropertyValueType.Text;
    }
    const type = entry.getAttribute(`${Entry_1.default.Attributes.FieldTypePrefix}${propertyName}`);
    return VALID_VALUE_TYPES.indexOf(type) >= 0 ? type : types_1.EntryPropertyValueType.Text;
}
exports.getEntryPropertyValueType = getEntryPropertyValueType;
/**
 * Check if an ID signifies a new instance and not an
 *  existing one
 * @param id The ID to check
 */
function idSignifiesNew(id, mergeMode = false) {
    if (/^\d+$/.test(id) === false)
        return mergeMode;
    const numerical = parseInt(id, 10);
    return !isNaN(numerical) && numerical > 0 && numerical <= 999999;
}
exports.idSignifiesNew = idSignifiesNew;
/**
 * Set the value type attribute of an entry
 * @param entry Entry instance
 * @param propertyName The property name
 * @param valueType The value type
 */
function setEntryPropertyValueType(entry, propertyName, valueType) {
    entry.setAttribute(`${Entry_1.default.Attributes.FieldTypePrefix}${propertyName}`, valueType);
}
exports.setEntryPropertyValueType = setEntryPropertyValueType;
