"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const layerr_1 = require("layerr");
const VaultFormat_1 = __importDefault(require("./VaultFormat"));
const Vault_1 = __importDefault(require("../core/Vault"));
const commands_1 = require("./formatA/commands");
const tools_1 = require("./formatA/tools");
const Flattener_1 = __importDefault(require("./formatA/Flattener"));
const signing_1 = require("./formatA/signing");
const describe_1 = require("./formatA/describe");
const VaultComparator_1 = __importDefault(require("./formatA/VaultComparator"));
const appEnv_1 = require("../env/appEnv");
const encoding_1 = require("../tools/encoding");
const uuid_1 = require("../tools/uuid");
const channel_1 = require("../credentials/channel");
const common_1 = require("./common");
const types_1 = require("../types");
const COMMANDS = {
    aid: commands_1.executeArchiveID,
    cen: commands_1.executeCreateEntry,
    cgr: commands_1.executeCreateGroup,
    cmm: commands_1.executeComment,
    daa: commands_1.executeDeleteArchiveAttribute,
    dea: commands_1.executeDeleteEntryAttribute,
    dem: commands_1.executeDeleteEntryProperty,
    den: commands_1.executeDeleteEntry,
    dep: commands_1.executeDeleteEntryProperty,
    dga: commands_1.executeDeleteGroupAttribute,
    dgr: commands_1.executeDeleteGroup,
    fmt: commands_1.executeFormat,
    men: commands_1.executeMoveEntry,
    mgr: commands_1.executeMoveGroup,
    pad: commands_1.executePad,
    saa: commands_1.executeSetArchiveAttribute,
    sea: commands_1.executeSetEntryAttribute,
    sem: commands_1.executeSetEntryProperty,
    sep: commands_1.executeSetEntryProperty,
    sga: commands_1.executeSetGroupAttribute,
    tgr: commands_1.executeTitleGroup
};
const SHARE_COMMAND_EXP = /^\$[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}\s/;
const UUID_LEN = 36;
const VALID_COMMAND_EXP = /^[a-z]{3}\s.+$/;
function emptyVault() {
    return {
        id: null,
        groups: []
    };
}
class VaultFormatA extends VaultFormat_1.default {
    static encodeRaw(rawContent, credentials) {
        const compress = (0, appEnv_1.getSharedAppEnv)().getProperty("compression/v1/compressText");
        const encrypt = (0, appEnv_1.getSharedAppEnv)().getProperty("crypto/v1/encryptText");
        const { masterPassword } = (0, channel_1.getCredentials)(credentials.id);
        return Promise.resolve()
            .then(() => (0, common_1.historyArrayToString)(rawContent))
            .then(history => compress(history))
            .then(compressed => encrypt(compressed, masterPassword))
            .then(signing_1.sign);
    }
    /**
     * Extract shares from a history collection
     * @param {String[]} history A history collection, containing shares
     * @returns {Object} The resulting separated histories. The object will
     *  always contain a `base` property containing the non-share history.
     *  Each share detected is set on the object under its share ID - being
     *  set to an array of history lines (non-prefixed) for that share.
     */
    static extractSharesFromHistory(history) {
        return history.reduce((output, line) => {
            if (SHARE_COMMAND_EXP.test(line)) {
                const shareID = line.substring(1, 1 + UUID_LEN);
                const command = line.replace(SHARE_COMMAND_EXP, "");
                output[shareID] = output[shareID] || [];
                output[shareID].push(command);
            }
            else {
                output.base.push(line);
            }
            return output;
        }, { base: [] });
    }
    static getFormatID() {
        return types_1.VaultFormatID.A;
    }
    static historiesDiffer(historyA, historyB) {
        const differences = VaultComparator_1.default.calculateHistoryDifferences(historyA, historyB);
        return differences === null || differences.original.length > 0 || differences.secondary.length > 0;
    }
    static isEncrypted(contents) {
        return (0, signing_1.vaultContentsEncrypted)(contents);
    }
    static parseEncrypted(encryptedContent, credentials) {
        const decompress = (0, appEnv_1.getSharedAppEnv)().getProperty("compression/v1/decompressText");
        const decrypt = (0, appEnv_1.getSharedAppEnv)().getProperty("crypto/v1/decryptText");
        const { masterPassword } = (0, channel_1.getCredentials)(credentials.id);
        return Promise.resolve()
            .then(() => {
            if (!(0, signing_1.hasValidSignature)(encryptedContent)) {
                throw new Error("No valid signature in vault");
            }
            return (0, signing_1.stripSignature)(encryptedContent);
        })
            .then(encryptedData => decrypt(encryptedData, masterPassword))
            .then(decrypted => {
            if (decrypted && decrypted.length > 0) {
                const decompressed = decompress(decrypted);
                if (decompressed) {
                    return (0, common_1.historyStringToArray)(decompressed, types_1.VaultFormatID.A);
                }
            }
            throw new Error("Failed reconstructing history: Decryption failed");
        });
    }
    static prepareHistoryForMerge(history) {
        return (0, tools_1.stripDestructiveCommands)(history);
    }
    static vaultFromMergedHistories(base, incoming) {
        const differences = VaultComparator_1.default.calculateHistoryDifferences(base, incoming);
        if (differences === null) {
            throw new Error("Failed merging vault histories: No common base");
        }
        if (differences.secondary.length === 0) {
            // Remote doesn't have unseen changes, so we can simply
            // use the existing history
            return Vault_1.default.createFromHistory(base, VaultFormatA);
        }
        // Remote has unseen changes, so we need to do a full merge
        // to manage the differences
        const newHistoryBase = VaultFormatA.prepareHistoryForMerge(differences.original);
        const newHistoryIncoming = VaultFormatA.prepareHistoryForMerge(differences.secondary);
        const common = differences.common;
        const newVault = new Vault_1.default(VaultFormatA);
        newVault.format.erase();
        // merge all history and execute on new vault
        newVault.format.execute(common.concat(newHistoryBase).concat(newHistoryIncoming));
        newVault.format.dirty = false;
        return newVault;
    }
    constructor(source = emptyVault()) {
        super(source);
        this._history = [];
        this._history.format = types_1.VaultFormatID.A;
    }
    cloneEntry(entry, targetGroupID) { }
    cloneGroup(group, targetGroupID) {
        const groupDesc = (0, describe_1.describeVaultDataset)(group, targetGroupID);
        this.execute(groupDesc);
    }
    createEntry(groupID, entryID) {
        this.execute(tools_1.InigoCommand.create(tools_1.InigoCommand.Command.CreateEntry)
            .addArgument(groupID)
            .addArgument(entryID)
            .generateCommand());
    }
    createGroup(parentID, groupID) {
        this.execute(tools_1.InigoCommand.create(tools_1.InigoCommand.Command.CreateGroup)
            .addArgument(parentID)
            .addArgument(groupID)
            .generateCommand());
    }
    deleteEntry(entryID) {
        this.execute(tools_1.InigoCommand.create(tools_1.InigoCommand.Command.DeleteEntry)
            .addArgument(entryID)
            .generateCommand());
    }
    deleteEntryAttribute(entryID, attribute) {
        this.execute(tools_1.InigoCommand.create(tools_1.InigoCommand.Command.DeleteEntryAttribute)
            .addArgument(entryID)
            .addArgument(attribute)
            .generateCommand());
    }
    deleteEntryProperty(entryID, property) {
        this.execute(tools_1.InigoCommand.create(tools_1.InigoCommand.Command.DeleteEntryProperty)
            .addArgument(entryID)
            .addArgument(property)
            .generateCommand());
    }
    deleteGroup(groupID) {
        this.execute(tools_1.InigoCommand.create(tools_1.InigoCommand.Command.DeleteGroup)
            .addArgument(groupID)
            .generateCommand());
    }
    deleteGroupAttribute(groupID, attribute) {
        this.execute(tools_1.InigoCommand.create(tools_1.InigoCommand.Command.DeleteGroupAttribute)
            .addArgument(groupID)
            .addArgument(attribute)
            .generateCommand());
    }
    deleteVaultAttribute(attribute) {
        this.execute(tools_1.InigoCommand.create(tools_1.InigoCommand.Command.DeleteArchiveAttribute)
            .addArgument(attribute)
            .generateCommand());
    }
    erase() {
        super.erase();
        Object.assign(this.source, emptyVault());
        this.emit("erased");
    }
    execute(commandOrCommands) {
        if (this.readOnly) {
            throw new Error("Format is in read-only mode");
        }
        const commands = Array.isArray(commandOrCommands) ? commandOrCommands : [commandOrCommands];
        commands.forEach(command => this._executeCommand(command));
        const lastCommand = commands[commands.length - 1];
        if (/^pad\s/i.test(lastCommand) === false) {
            this._pad();
        }
        this.dirty = true;
        this.emit("commandsExecuted");
    }
    findEntryByID(id) {
        return this.getAllEntries().find(entry => entry.id === id) || null;
    }
    findGroupByID(id) {
        return this.getAllGroups().find(group => group.id === id) || null;
    }
    findGroupContainingEntryID(id) {
        const matchingEntry = this.getAllEntries().find(entry => entry.id === id);
        if (matchingEntry) {
            return this.getAllGroups().find(group => group.id === matchingEntry.parentID) || null;
        }
        return null;
    }
    findGroupContainingGroupID(id) {
        const searchGroups = (groups) => {
            for (const group of groups) {
                if (group.id === id)
                    return null;
                const children = group.groups || [];
                const childMatch = !!children.find(child => child.id === id);
                if (childMatch)
                    return group;
                const deepGroup = searchGroups(children);
                if (deepGroup)
                    return deepGroup;
            }
            return null;
        };
        return searchGroups(this.getAllGroups());
    }
    generateID() {
        this.execute(tools_1.InigoCommand.create(tools_1.InigoCommand.Command.ArchiveID)
            .addArgument((0, uuid_1.generateUUID)())
            .generateCommand());
    }
    getAllEntries(parentID = null) {
        return (0, tools_1.getAllEntries)(this.source, parentID);
    }
    getAllGroups(parentID = null) {
        return (0, tools_1.getAllGroups)(this.source, parentID);
    }
    getEntryAttributes(entrySource) {
        return entrySource.attributes;
    }
    getEntryChanges(entrySource) {
        return (entrySource.history || [])
            .filter(item => item.propertyType === types_1.EntryPropertyType.Property)
            .map(item => {
            const type = !item.originalValue
                ? types_1.EntryChangeType.Created
                : typeof item.newValue === "string"
                    ? types_1.EntryChangeType.Modified
                    : types_1.EntryChangeType.Deleted;
            const change = {
                property: item.property,
                type,
                ts: null
            };
            if (type === types_1.EntryChangeType.Created || types_1.EntryChangeType.Modified) {
                change.value = item.newValue;
            }
            return change;
        });
    }
    getEntryProperties(entrySource) {
        return entrySource.properties;
    }
    getFormat() {
        return VaultFormatA;
    }
    getGroupAttributes(groupSource) {
        return groupSource.attributes;
    }
    getGroupTitle(groupSource) {
        return groupSource.title;
    }
    getHistory() {
        if (!this._history.format) {
            this._history.format = types_1.VaultFormatID.A;
        }
        return this._history;
    }
    getItemID(itemSource) {
        return itemSource.id;
    }
    getItemParentID(itemSource) {
        return itemSource.parentID;
    }
    getVaultAttributes() {
        return this.source.attributes || {};
    }
    getVaultID() {
        return this.source.id;
    }
    initialise() {
        this.execute(tools_1.InigoCommand.create(tools_1.InigoCommand.Command.Format)
            .addArgument((0, signing_1.getFormat)())
            .generateCommand());
        this.generateID();
    }
    moveEntry(entryID, groupID) {
        this.execute(tools_1.InigoCommand.create(tools_1.InigoCommand.Command.MoveEntry)
            .addArgument(entryID)
            .addArgument(groupID)
            .generateCommand());
    }
    moveGroup(groupID, newParentID) {
        this.execute(tools_1.InigoCommand.create(tools_1.InigoCommand.Command.MoveGroup)
            .addArgument(groupID)
            .addArgument(newParentID)
            .generateCommand());
    }
    optimise() {
        const flattener = new Flattener_1.default(this);
        if (flattener.canBeFlattened()) {
            flattener.flatten();
        }
    }
    setEntryAttribute(entryID, attribute, value) {
        this.execute(tools_1.InigoCommand.create(tools_1.InigoCommand.Command.SetEntryAttribute)
            .addArgument(entryID)
            .addArgument(attribute)
            .addArgument(value)
            .generateCommand());
    }
    setEntryProperty(entryID, property, value) {
        this.execute(tools_1.InigoCommand.create(tools_1.InigoCommand.Command.SetEntryProperty)
            .addArgument(entryID)
            .addArgument(property)
            .addArgument(value)
            .generateCommand());
    }
    setGroupAttribute(groupID, attribute, value) {
        this.execute(tools_1.InigoCommand.create(tools_1.InigoCommand.Command.SetGroupAttribute)
            .addArgument(groupID)
            .addArgument(attribute)
            .addArgument(value)
            .generateCommand());
    }
    setGroupTitle(groupID, title) {
        this.execute(tools_1.InigoCommand.create(tools_1.InigoCommand.Command.SetGroupTitle)
            .addArgument(groupID)
            .addArgument(title)
            .generateCommand());
    }
    setVaultAttribute(key, value) {
        this.execute(tools_1.InigoCommand.create(tools_1.InigoCommand.Command.SetArchiveAttribute)
            .addArgument(key)
            .addArgument(value)
            .generateCommand());
    }
    _executeCommand(command) {
        let currentCommand = command, shareID = null;
        if (SHARE_COMMAND_EXP.test(currentCommand)) {
            const shareMatch = /^\$([a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12})/.exec(currentCommand);
            shareID = shareMatch[1];
            currentCommand = currentCommand.replace(SHARE_COMMAND_EXP, "");
        }
        if (!VALID_COMMAND_EXP.test(currentCommand)) {
            throw new Error(`Invalid command: ${command}`);
        }
        if (/^pad\s/i.test(command) && /^pad\s/i.test(this.history[this.history.length - 1])) {
            // Skip adding extra pad
            return;
        }
        const commandComponents = (0, tools_1.extractCommandComponents)(currentCommand);
        const commandKey = commandComponents.shift().toLowerCase();
        const executeCommand = COMMANDS[commandKey];
        try {
            executeCommand.apply(null, [
                this.source,
                {
                    // opts
                    shareID
                },
                ...this._processCommandParameters(commandKey, commandComponents)
            ]);
            this.history.push(command);
        }
        catch (err) {
            throw new layerr_1.Layerr(err, `Failed executing vault command: ${commandKey}`);
        }
    }
    _pad() {
        this._executeCommand(tools_1.InigoCommand.generatePaddingCommand());
    }
    _processCommandParameters(commandKey, parameters) {
        const friendlyCommand = Object.keys(tools_1.COMMAND_MANIFEST).find(manifestKey => {
            return tools_1.COMMAND_MANIFEST[manifestKey].s === commandKey;
        });
        const commandDescriptor = tools_1.COMMAND_MANIFEST[friendlyCommand];
        if (!commandDescriptor) {
            throw new Error(`Cannot process command parameters: no command found for key: ${commandKey}`);
        }
        return parameters.map((parameter, i) => {
            if (commandDescriptor.args[i].encode === true) {
                if ((0, encoding_1.isEncoded)(parameter)) {
                    return (0, encoding_1.decodeStringValue)(parameter);
                }
            }
            return parameter;
        });
    }
}
exports.default = VaultFormatA;
