"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.BaseSearch = void 0;
const fast_levenshtein_1 = __importDefault(require("fast-levenshtein"));
const StorageInterface_1 = __importDefault(require("../storage/StorageInterface"));
const searcher_1 = require("./searcher");
function domainsRelated(domain1, domain2) {
    if (domain1 === domain2)
        return true;
    if (domain1.length > domain2.length) {
        const ind = domain1.indexOf(domain2);
        return ind === domain1.length - domain2.length;
    }
    else {
        const ind = domain2.indexOf(domain1);
        return ind === domain2.length - domain1.length;
    }
}
function extractDomain(str) {
    const match = /^((https?|ftp):\/\/)?([^\/]+)/i.exec(str);
    return (match && match[3]) || "";
}
class BaseSearch {
    constructor(targets, entryFetcher, memory = new StorageInterface_1.default(), searcherFactory = searcher_1.buildSearcher) {
        this._entries = [];
        this._fuse = null;
        this._results = [];
        this._scores = {};
        this._targets = targets;
        this._entryFetcher = entryFetcher;
        this._memory = memory;
        this._searcherFactory = searcherFactory;
    }
    /**
     * Last search results
     */
    get results() {
        return this._results;
    }
    /**
     * Increment the score of a URL in an entry
     * @param vaultID The vault ID
     * @param entryID The entry ID
     * @param url The URL to increment for
     */
    incrementScore(vaultID, entryID, url) {
        return __awaiter(this, void 0, void 0, function* () {
            const scoresRaw = yield this._memory.getValue(`bcup_search_${vaultID}`);
            let vaultScore = {};
            if (scoresRaw) {
                try {
                    vaultScore = JSON.parse(scoresRaw);
                }
                catch (err) { }
            }
            const domain = extractDomain(url);
            if (!domain)
                return;
            vaultScore[entryID] = vaultScore[entryID] || {};
            vaultScore[entryID][domain] = vaultScore[entryID][domain] ? vaultScore[entryID][domain] + 1 : 1;
            yield this._memory.setValue(`bcup_search_${vaultID}`, JSON.stringify(vaultScore));
        });
    }
    /**
     * Prepare the search instance by processing
     * entries
     */
    prepare() {
        return __awaiter(this, void 0, void 0, function* () {
            this._entries = [];
            this._scores = {};
            for (const target of this._targets) {
                // Get scores
                const scoresRaw = yield this._memory.getValue(`bcup_search_${target.id}`);
                if (scoresRaw) {
                    try {
                        const scores = JSON.parse(scoresRaw);
                        this._scores[target.id] = scores;
                    }
                    catch (err) { }
                }
                // Process entries
                const entries = yield this._entryFetcher(target, this._memory);
                this._entries.push(...entries);
            }
            // Instantiate new searcher
            this._fuse = this._searcherFactory(this._entries);
        });
    }
    /**
     * Search for entries by term
     * @param term The term to search for
     * @returns An array of search results
     */
    searchByTerm(term) {
        if (!this._fuse) {
            throw new Error("Searching interface not prepared");
        }
        this._results = this._fuse.search(term).map(result => result.item);
        return this._results;
    }
    /**
     * Search for entries by URL
     * @param url The URL to search with
     * @returns An array of search results
     */
    searchByURL(url) {
        const incomingDomain = extractDomain(url);
        if (!incomingDomain) {
            this._results = [];
            return this._results;
        }
        const results = this._entries.reduce((output, entryItem) => {
            let bestScore = Infinity;
            const bestURL = entryItem.urls.reduce((best, next) => {
                const thisDomain = extractDomain(next);
                if (!thisDomain)
                    return best;
                if (!domainsRelated(thisDomain, incomingDomain))
                    return best;
                const score = fast_levenshtein_1.default.get(next, url);
                if (score < bestScore) {
                    bestScore = score;
                    return next;
                }
                return best;
            }, null);
            if (!bestURL)
                return output;
            const resultDomain = extractDomain(bestURL);
            return [
                ...output,
                {
                    item: entryItem,
                    score: bestScore,
                    url: bestURL,
                    domainScore: Math.max(entryItem.domainScores[incomingDomain] || 0, entryItem.domainScores[resultDomain] || 0)
                }
            ];
        }, []);
        this._results = results
            .sort((a, b) => {
            if (a.domainScore > b.domainScore) {
                return -1;
            }
            else if (b.domainScore > a.domainScore) {
                return 1;
            }
            if (a.score > b.score) {
                return 1;
            }
            else if (b.score > a.score) {
                return -1;
            }
            return 0;
        })
            .map(result => result.item);
        return this._results;
    }
}
exports.BaseSearch = BaseSearch;
