/// <reference types="node" />
import Credentials from "../credentials/Credentials";
import Entry from "../core/Entry";
import VaultSource from "../core/VaultSource";
import { AttachmentDetails } from "../types";
/**
 * Attachment manager
 * @memberof module:Buttercup
 */
export default class AttachmentManager {
    /**
     * Get a new attachment ID
     * @static
     * @memberof AttachmentManager
     */
    static newAttachmentID(): string;
    _credentials: Credentials;
    _source: VaultSource;
    /**
     * Constructor for new attachment managers
     * @param vaultSource The vault source to attach to. This is
     *  normally set by the VaultSource automatically when unlocking a source.
     */
    constructor(vaultSource: VaultSource);
    /**
     * Get an attachment's data
     * @param entry Entry instance the attachment can be found on
     * @param attachmentID The attachment ID
     * @throws {Error} Throws if the attachment isn't found
     * @memberof AttachmentManager
     */
    getAttachment(entry: Entry, attachmentID: string): Promise<Buffer | ArrayBuffer>;
    /**
     * Get an attachment's details
     * @param entry Entry instance the attachment can be found on
     * @param attachmentID The attachment ID
     * @returns The details or null if the attachment isn't found
     * @memberof AttachmentManager
     */
    getAttachmentDetails(entry: Entry, attachmentID: string): Promise<AttachmentDetails>;
    /**
     * List all attachments
     * @param entry Entry instance the attachment can be found on
     * @memberof AttachmentManager
     */
    listAttachments(entry: Entry): Promise<Array<AttachmentDetails>>;
    /**
     * Remove an attachment, deleting the file and removing it from
     *  the entry
     * @param entry Entry instance the attachment can be found on
     * @param attachmentID The attachment ID
     * @memberof AttachmentManager
     */
    removeAttachment(entry: Entry, attachmentID: string): Promise<void>;
    /**
     * Write an attachment to an entry.
     * Creates a new attachment or updates an existing one, by attachment ID
     * @param entry Entry instance the attachment can be found on
     * @param attachmentID The attachment ID
     * @param attachmentData The attachment's data
     * @param name The file name
     * @param type The MIME type
     * @param timestamp Optional timestamp override for the creation of the
     *  attachment
     * @memberof AttachmentManager
     */
    setAttachment(entry: Entry, attachmentID: string, attachmentData: Buffer | ArrayBuffer, name: string, type: string, timestamp?: Date): Promise<void>;
    _checkAttachmentSupport(): void;
    _getAttachmentsCredentials(): Promise<Credentials>;
}
